@@ -26,20 +26,16 @@ WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.  */
 #ifndef dwarf_h
 #define dwarf_h
 
-#ifdef HAVE_CONFIG_H
-#include <config.h>
-#endif
-
-#ifdef HAVE_ATOMIC_OPS_H
-# include <atomic_ops.h>
-#endif
-
 #include <libunwind.h>
 
 struct dwarf_cursor;	/* forward-declaration */
+struct elf_dyn_info;
 
 #include "dwarf-config.h"
+#ifndef UNW_REMOTE_ONLY
 #include <link.h>
+#endif
+#include <pthread.h>
 
 /* DWARF expression opcodes.  */
 
@@ -248,7 +244,8 @@ typedef struct dwarf_reg_state
     unsigned short lru_chain;	  /* used for least-recently-used chain */
     unsigned short coll_chain;	/* used for hash collisions */
     unsigned short hint;	      /* hint for next rs to try (or -1) */
-    unsigned short signal_frame; /* optional machine-dependent signal info */
+    unsigned short valid : 1;         /* optional machine-dependent signal info */
+    unsigned short signal_frame : 1;  /* optional machine-dependent signal info */
   }
 dwarf_reg_state_t;
 
@@ -365,6 +362,7 @@ struct dwarf_callback_data
 #define dwarf_find_proc_info		UNW_OBJ (dwarf_find_proc_info)
 #define dwarf_find_debug_frame		UNW_OBJ (dwarf_find_debug_frame)
 #define dwarf_search_unwind_table	UNW_OBJ (dwarf_search_unwind_table)
+#define dwarf_find_unwind_table		UNW_OBJ (dwarf_find_unwind_table)
 #define dwarf_put_unwind_info		UNW_OBJ (dwarf_put_unwind_info)
 #define dwarf_put_unwind_info		UNW_OBJ (dwarf_put_unwind_info)
 #define dwarf_eval_expr			UNW_OBJ (dwarf_eval_expr)
@@ -377,17 +375,24 @@ struct dwarf_callback_data
 #define dwarf_step			UNW_OBJ (dwarf_step)
 
 extern int dwarf_init (void);
+#ifndef UNW_REMOTE_ONLY
 extern int dwarf_callback (struct dl_phdr_info *info, size_t size, void *ptr);
 extern int dwarf_find_proc_info (unw_addr_space_t as, unw_word_t ip,
 				 unw_proc_info_t *pi,
 				 int need_unwind_info, void *arg);
+#endif /* !UNW_REMOTE_ONLY */
 extern int dwarf_find_debug_frame (int found, unw_dyn_info_t *di_debug,
-				 struct dl_phdr_info *info, unw_word_t ip);
+				   unw_word_t ip, unw_word_t segbase,
+				   const char* obj_name, unw_word_t start,
+				   unw_word_t end);
 extern int dwarf_search_unwind_table (unw_addr_space_t as,
 				      unw_word_t ip,
 				      unw_dyn_info_t *di,
 				      unw_proc_info_t *pi,
 				      int need_unwind_info, void *arg);
+extern int dwarf_find_unwind_table (struct elf_dyn_info *edi, unw_addr_space_t as,
+				    char *path, unw_word_t segbase, unw_word_t mapoff,
+				    unw_word_t ip);
 extern void dwarf_put_unwind_info (unw_addr_space_t as,
 				   unw_proc_info_t *pi, void *arg);
 extern int dwarf_eval_expr (struct dwarf_cursor *c, unw_word_t *addr,
