@@ -35,6 +35,8 @@ WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.  */
 # include "config.h"
 #endif
 
+#include "compiler.h"
+
 #ifdef HAVE___THREAD
   /* For now, turn off per-thread caching.  It uses up too much TLS
      memory per thread even when the thread never uses libunwind at
@@ -53,6 +55,8 @@ WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.  */
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
+#include <sys/mman.h>
+#include <elf.h>
 
 #if defined(HAVE_ENDIAN_H)
 # include <endian.h>
@@ -68,45 +72,12 @@ WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.  */
 # endif
 #endif
 
-#ifdef __GNUC__
-# define UNUSED		__attribute__((unused))
-# define NORETURN	__attribute__((noreturn))
-# define ALIAS(name)	__attribute__((alias (#name)))
-# if (__GNUC__ > 3) || (__GNUC__ == 3 && __GNUC_MINOR__ > 2)
-#  define ALWAYS_INLINE	inline __attribute__((always_inline))
-#  define HIDDEN	__attribute__((visibility ("hidden")))
-#  define PROTECTED	__attribute__((visibility ("protected")))
-# else
-#  define ALWAYS_INLINE
-#  define HIDDEN
-#  define PROTECTED
-# endif
-# if (__GNUC__ >= 3)
-#  define likely(x)	__builtin_expect ((x), 1)
-#  define unlikely(x)	__builtin_expect ((x), 0)
-# else
-#  define likely(x)	(x)
-#  define unlikely(x)	(x)
-# endif
-#else
-# define ALWAYS_INLINE
-# define UNUSED
-# define NORETURN
-# define ALIAS(name)
-# define HIDDEN
-# define PROTECTED
-# define likely(x)	(x)
-# define unlikely(x)	(x)
-#endif
-
 #ifdef DEBUG
 # define UNW_DEBUG	1
 #else
 # define UNW_DEBUG	0
 #endif
 
-#define ARRAY_SIZE(a)	(sizeof (a) / sizeof ((a)[0]))
-
 /* Make it easy to write thread-safe code which may or may not be
    linked against libpthread.  The macros below can be used
    unconditionally and if -lpthread is around, they'll call the
@@ -117,11 +88,11 @@ WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.  */
 #pragma weak pthread_mutex_unlock
 
 #define mutex_init(l)							\
-	(pthread_mutex_init != 0 ? pthread_mutex_init ((l), 0) : 0)
+	(pthread_mutex_init != NULL ? pthread_mutex_init ((l), NULL) : 0)
 #define mutex_lock(l)							\
-	(pthread_mutex_lock != 0 ? pthread_mutex_lock (l) : 0)
+	(pthread_mutex_lock != NULL ? pthread_mutex_lock (l) : 0)
 #define mutex_unlock(l)							\
-	(pthread_mutex_unlock != 0 ? pthread_mutex_unlock (l) : 0)
+	(pthread_mutex_unlock != NULL ? pthread_mutex_unlock (l) : 0)
 
 #ifdef HAVE_ATOMIC_OPS_H
 # include <atomic_ops.h>
@@ -139,14 +110,16 @@ cmpxchg_ptr (void *addr, void *old, void *new)
   return AO_compare_and_swap(u.aop, (AO_t) old, (AO_t) new);
 }
 # define fetch_and_add1(_ptr)		AO_fetch_and_add1(_ptr)
+# define fetch_and_add(_ptr, value)	AO_fetch_and_add(_ptr, value)
    /* GCC 3.2.0 on HP-UX crashes on cmpxchg_ptr() */
 #  if !(defined(__hpux) && __GNUC__ == 3 && __GNUC_MINOR__ == 2)
 #   define HAVE_CMPXCHG
 #  endif
-# define HAVE_FETCH_AND_ADD1
-#else
+# define HAVE_FETCH_AND_ADD
+#elif defined(HAVE_SYNC_ATOMICS) || defined(HAVE_IA64INTRIN_H)
 # ifdef HAVE_IA64INTRIN_H
 #  include <ia64intrin.h>
+# endif
 static inline int
 cmpxchg_ptr (void *addr, void *old, void *new)
 {
@@ -160,10 +133,10 @@ cmpxchg_ptr (void *addr, void *old, void *new)
   u.vp = addr;
   return __sync_bool_compare_and_swap(u.vlp, (long) old, (long) new);
 }
-#  define fetch_and_add1(_ptr)		__sync_fetch_and_add(_ptr, 1)
-#  define HAVE_CMPXCHG
-#  define HAVE_FETCH_AND_ADD1
-# endif
+# define fetch_and_add1(_ptr)		__sync_fetch_and_add(_ptr, 1)
+# define fetch_and_add(_ptr, value)	__sync_fetch_and_add(_ptr, value)
+# define HAVE_CMPXCHG
+# define HAVE_FETCH_AND_ADD
 #endif
 #define atomic_read(ptr)	(*(ptr))
 
@@ -182,7 +155,7 @@ extern intrmask_t unwi_full_mask;
 
 /* Silence compiler warnings about variables which are used only if libunwind
    is configured in a certain way */
-static inline void mark_as_used(void *v) {
+static inline void mark_as_used(void *v UNUSED) {
 }
 
 #if defined(CONFIG_BLOCK_SIGNALS)
@@ -214,8 +187,8 @@ do {						\
 #define GET_MEMORY(mem, size)				    		    \
 do {									    \
   /* Hopefully, mmap() goes straight through to a system call stub...  */   \
-  mem = mmap (0, size, PROT_READ | PROT_WRITE, MAP_PRIVATE | MAP_ANONYMOUS, \
-	      -1, 0);							    \
+  mem = mmap (NULL, size, PROT_READ | PROT_WRITE,			    \
+	      MAP_PRIVATE | MAP_ANONYMOUS, -1, 0);			    \
   if (mem == MAP_FAILED)						    \
     mem = NULL;								    \
 } while (0)
@@ -301,6 +274,32 @@ struct elf_image
     size_t size;		/* (file-) size of the image */
   };
 
+struct elf_dyn_info
+  {
+    struct elf_image ei;
+    unw_dyn_info_t di_cache;
+    unw_dyn_info_t di_debug;    /* additional table info for .debug_frame */
+#if UNW_TARGET_IA64
+    unw_dyn_info_t ktab;
+#endif
+#if UNW_TARGET_ARM
+    unw_dyn_info_t di_arm;      /* additional table info for .ARM.exidx */
+#endif
+  };
+
+static inline void invalidate_edi (struct elf_dyn_info *edi)
+{
+  if (edi->ei.image)
+    munmap (edi->ei.image, edi->ei.size);
+  memset (edi, 0, sizeof (*edi));
+  edi->di_cache.format = -1;
+  edi->di_debug.format = -1;
+#if UNW_TARGET_ARM
+  edi->di_arm.format = -1;
+#endif
+}
+
+
 /* Provide a place holder for architecture to override for fast access
    to memory when known not to need to validate and know the access
    will be local to the process. A suitable override will improve
@@ -309,10 +308,21 @@ struct elf_image
   do { (ret) = dwarf_get ((cur), DWARF_MEM_LOC ((cur), (addr)), &(to)); } \
   while (0)
 
+/* Define GNU and processor specific values for the Phdr p_type field in case
+   they aren't defined by <elf.h>.  */
+#ifndef PT_GNU_EH_FRAME
+# define PT_GNU_EH_FRAME	0x6474e550
+#endif /* !PT_GNU_EH_FRAME */
+#ifndef PT_ARM_EXIDX
+# define PT_ARM_EXIDX		0x70000001	/* ARM unwind segment */
+#endif /* !PT_ARM_EXIDX */
+
 #include "tdep/libunwind_i.h"
 
 #ifndef tdep_get_func_addr
 # define tdep_get_func_addr(as,addr,v)		(*(v) = addr, 0)
 #endif
 
+#define UNW_ALIGN(x,a) (((x)+(a)-1UL)&~((a)-1UL))
+
 #endif /* libunwind_i_h */
