@@ -1,6 +1,7 @@
 /* libunwind - a platform-independent unwind library
    Copyright (C) 2008 CodeSourcery
    Copyright 2011 Linaro Limited
+   Copyright (C) 2012 Tommi Rantala <tt.rantala@gmail.com>
 
 This file is part of libunwind.
 
@@ -67,6 +68,8 @@ arm_exidx_step (struct cursor *c)
       return -UNW_EBADFRAME;
     }
 
+  c->dwarf.pi_valid = 0;
+
   return (c->dwarf.ip == 0) ? 0 : 1;
 }
 
@@ -111,7 +114,6 @@ unw_handle_signal_frame (unw_cursor_t *cursor)
 	  c->sigcontext_format = ARM_SCF_LINUX_OLD_SIGFRAME;
 	  sc_addr = sp_addr;
 	}
-      c->sigcontext_addr = sp_addr;
     }
   else if (ret == 2)
     {
@@ -120,18 +122,19 @@ unw_handle_signal_frame (unw_cursor_t *cursor)
       if (sp == sp_addr + 8)
 	{
 	  c->sigcontext_format = ARM_SCF_LINUX_OLD_RT_SIGFRAME;
-	  c->sigcontext_addr = sp_addr + 8 + sizeof (siginfo_t); 
+	  sc_addr = sp_addr + 8 + sizeof (siginfo_t) + LINUX_UC_MCONTEXT_OFF;
 	}
       else
 	{
 	  c->sigcontext_format = ARM_SCF_LINUX_RT_SIGFRAME;
-	  c->sigcontext_addr = sp_addr + sizeof (siginfo_t);
+	  sc_addr = sp_addr + sizeof (siginfo_t) + LINUX_UC_MCONTEXT_OFF;
 	}
-      sc_addr = c->sigcontext_addr + LINUX_UC_MCONTEXT_OFF;
     }
   else
     return -UNW_EUNSPEC;
 
+  c->sigcontext_addr = sc_addr;
+
   /* Update the dwarf cursor.
      Set the location of the registers to the corresponding addresses of the
      uc_mcontext / sigcontext structure contents.  */
@@ -156,6 +159,8 @@ unw_handle_signal_frame (unw_cursor_t *cursor)
   dwarf_get (&c->dwarf, c->dwarf.loc[UNW_ARM_R13], &c->dwarf.cfa);
   dwarf_get (&c->dwarf, c->dwarf.loc[UNW_ARM_R15], &c->dwarf.ip);
 
+  c->dwarf.pi_valid = 0;
+
   return 1;
 }
 
@@ -171,6 +176,7 @@ unw_step (unw_cursor_t *cursor)
   if (unw_is_signal_frame (cursor))
      return unw_handle_signal_frame (cursor);
 
+#ifdef CONFIG_DEBUG_FRAME
   /* First, try DWARF-based unwinding. */
   if (UNW_TRY_METHOD(UNW_ARM_METHOD_DWARF))
     {
@@ -188,6 +194,7 @@ unw_step (unw_cursor_t *cursor)
         return ret;
       }
     }
+#endif /* CONFIG_DEBUG_FRAME */
 
   /* Next, try extbl-based unwinding. */
   if (UNW_TRY_METHOD (UNW_ARM_METHOD_EXIDX))
@@ -250,6 +257,7 @@ unw_step (unw_cursor_t *cursor)
                 }
               c->dwarf.loc[UNW_ARM_R12] = ip_loc;
               c->dwarf.loc[UNW_ARM_R11] = fp_loc;
+              c->dwarf.pi_valid = 0;
               Debug(15, "ip=%lx\n", c->dwarf.ip);
             }
           else
