@@ -26,7 +26,9 @@ WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.  */
 /* Locate an FDE via the ELF data-structures defined by LSB v1.3
    (http://www.linuxbase.org/spec/).  */
 
+#ifndef UNW_REMOTE_ONLY
 #include <link.h>
+#endif /* !UNW_REMOTE_ONLY */
 #include <stddef.h>
 #include <stdio.h>
 #include <limits.h>
@@ -79,6 +81,7 @@ linear_search (unw_addr_space_t as, unw_word_t ip,
     }
   return -UNW_ENOINFO;
 }
+#endif /* !UNW_REMOTE_ONLY */
 
 #ifdef CONFIG_DEBUG_FRAME
 /* Load .debug_frame section from FILE.  Allocates and returns space
@@ -223,9 +226,9 @@ load_debug_frame (const char *file, char **buf, size_t *bufsize, int is_local)
 
 /* An error reading image file. Release resources and return error code */
 file_error:
-  if (stringtab) free(stringtab);
-  if (sec_hdrs) free(sec_hdrs);
-  if (linkbuf) free(linkbuf);
+  free(stringtab);
+  free(sec_hdrs);
+  free(linkbuf);
   fclose(f);
 
   return 1;
@@ -238,7 +241,7 @@ file_error:
 static int
 find_binary_for_address (unw_word_t ip, char *name, size_t name_size)
 {
-#ifdef __linux
+#if defined(__linux) && (!UNW_REMOTE_ONLY)
   struct map_iterator mi;
   int found = 0;
   int pid = getpid ();
@@ -268,17 +271,15 @@ find_binary_for_address (unw_word_t ip, char *name, size_t name_size)
    pointer to debug frame descriptor, or zero if not found.  */
 
 static struct unw_debug_frame_list *
-locate_debug_info (unw_addr_space_t as, struct dl_phdr_info *info,
-		   unw_word_t addr, const char *dlname)
+locate_debug_info (unw_addr_space_t as, unw_word_t addr, const char *dlname,
+		   unw_word_t start, unw_word_t end)
 {
   struct unw_debug_frame_list *w, *fdesc = 0;
   char path[PATH_MAX];
   char *name = path;
   int err;
-  uint64_t start = 0, end = 0;
   char *buf;
   size_t bufsize;
-  unsigned int i;
 
   /* First, see if we loaded this frame already.  */
 
@@ -305,29 +306,6 @@ locate_debug_info (unw_addr_space_t as, struct dl_phdr_info *info,
   else
     name = (char*) dlname;
 
-  /* Find the start/end of the described region by parsing the
-     dl_phdr_info structure.  */
-
-  start = info->dlpi_addr + info->dlpi_phdr[0].p_vaddr;
-  end = start;
-
-  for (i = 0; i < info->dlpi_phnum; i++)
-    {
-      Elf_W (Addr) hdrbase = info->dlpi_addr + info->dlpi_phdr[i].p_vaddr;
-      Elf_W (Addr) hdrlimit = hdrbase + info->dlpi_phdr[i].p_memsz;
-
-      if (info->dlpi_phdr[i].p_type != PT_LOAD)
-	continue;
-
-      if (hdrbase < start)
-	start = hdrbase;
-      if (hdrlimit > end)
-	end = hdrlimit;
-    }
-
-  Debug (4, "calculated bounds of %lx-%lx for '%s'\n", (long)start, (long)end,
-	 name);
-  
   err = load_debug_frame (name, &buf, &bufsize, as == unw_local_addr_space);
   
   if (!err)
@@ -354,18 +332,6 @@ struct debug_frame_tab
     uint32_t size;
   };
 
-static struct debug_frame_tab *
-debug_frame_tab_new (unsigned int base_size)
-{
-  struct debug_frame_tab *tab = malloc (sizeof (struct debug_frame_tab));
-
-  tab->tab = calloc (base_size, sizeof (struct table_entry));
-  tab->length = 0;
-  tab->size = base_size;
-  
-  return tab;
-}
-
 static void
 debug_frame_tab_append (struct debug_frame_tab *tab,
 			unw_word_t fde_offset, unw_word_t start_ip)
@@ -408,19 +374,19 @@ debug_frame_tab_compare (const void *a, const void *b)
 }
 
 PROTECTED int
-dwarf_find_debug_frame (int found, unw_dyn_info_t *di_debug,
-                       struct dl_phdr_info *info, unw_word_t ip)
+dwarf_find_debug_frame (int found, unw_dyn_info_t *di_debug, unw_word_t ip,
+			unw_word_t segbase, const char* obj_name,
+			unw_word_t start, unw_word_t end)
 {
   unw_dyn_info_t *di;
   struct unw_debug_frame_list *fdesc = 0;
   unw_accessors_t *a;
   unw_word_t addr;
 
-  Debug (15, "Trying to find .debug_frame info->dlpi_name=%s\n",
-        info->dlpi_name);
+  Debug (15, "Trying to find .debug_frame for %s\n", obj_name);
   di = di_debug;
 
-  fdesc = locate_debug_info (unw_local_addr_space, info, ip, info->dlpi_name);
+  fdesc = locate_debug_info (unw_local_addr_space, ip, obj_name, start, end);
 
   if (!fdesc)
     {
@@ -434,7 +400,7 @@ dwarf_find_debug_frame (int found, unw_dyn_info_t *di_debug,
       unw_word_t item_start, item_end = 0;
       uint32_t u32val = 0;
       uint64_t cie_id = 0;
-      struct debug_frame_tab *tab;
+      struct debug_frame_tab tab;
 
       Debug (15, "loaded .debug_frame\n");
 
@@ -457,7 +423,9 @@ dwarf_find_debug_frame (int found, unw_dyn_info_t *di_debug,
          /* Find all FDE entries in debug_frame, and make into a sorted
             index.  */
 
-         tab = debug_frame_tab_new (16);
+         tab.length = 0;
+         tab.size = 16;
+         tab.tab = calloc (tab.size, sizeof (struct table_entry));
 
          while (addr < (unw_word_t) (uintptr_t) (buf + bufsize))
            {
@@ -510,7 +478,7 @@ dwarf_find_debug_frame (int found, unw_dyn_info_t *di_debug,
                    {
                      Debug (15, "start_ip = %lx, end_ip = %lx\n",
                             (long) this_pi.start_ip, (long) this_pi.end_ip);
-                     debug_frame_tab_append (tab,
+                     debug_frame_tab_append (&tab,
                                              item_start - (unw_word_t) (uintptr_t) buf,
                                              this_pi.start_ip);
                    }
@@ -521,27 +489,26 @@ dwarf_find_debug_frame (int found, unw_dyn_info_t *di_debug,
              addr = item_end;
            }
 
-         debug_frame_tab_shrink (tab);
-         qsort (tab->tab, tab->length, sizeof (struct table_entry),
+         debug_frame_tab_shrink (&tab);
+         qsort (tab.tab, tab.length, sizeof (struct table_entry),
                 debug_frame_tab_compare);
-         /* for (i = 0; i < tab->length; i++)
+         /* for (i = 0; i < tab.length; i++)
             {
             fprintf (stderr, "ip %x, fde offset %x\n",
-            (int) tab->tab[i].start_ip_offset,
-            (int) tab->tab[i].fde_offset);
+            (int) tab.tab[i].start_ip_offset,
+            (int) tab.tab[i].fde_offset);
             }*/
-         fdesc->index = tab->tab;
-         fdesc->index_size = tab->length;
-         free (tab);
+         fdesc->index = tab.tab;
+         fdesc->index_size = tab.length;
        }
 
       di->format = UNW_INFO_FORMAT_TABLE;
       di->start_ip = fdesc->start;
       di->end_ip = fdesc->end;
-      di->u.ti.name_ptr = (unw_word_t) (uintptr_t) info->dlpi_name;
+      di->u.ti.name_ptr = (unw_word_t) (uintptr_t) obj_name;
       di->u.ti.table_data = (unw_word_t *) fdesc;
       di->u.ti.table_len = sizeof (*fdesc) / sizeof (unw_word_t);
-      di->u.ti.segbase = (unw_word_t) (uintptr_t) info->dlpi_addr;
+      di->u.ti.segbase = segbase;
 
       found = 1;
       Debug (15, "found debug_frame table `%s': segbase=0x%lx, len=%lu, "
@@ -555,6 +522,8 @@ dwarf_find_debug_frame (int found, unw_dyn_info_t *di_debug,
 
 #endif /* CONFIG_DEBUG_FRAME */
 
+#ifndef UNW_REMOTE_ONLY
+
 /* ptr is a pointer to a dwarf_callback_data structure and, on entry,
    member ip contains the instruction-pointer we're looking
    for.  */
@@ -565,13 +534,16 @@ dwarf_callback (struct dl_phdr_info *info, size_t size, void *ptr)
   unw_dyn_info_t *di = &cb_data->di;
   const Elf_W(Phdr) *phdr, *p_eh_hdr, *p_dynamic, *p_text;
   unw_word_t addr, eh_frame_start, eh_frame_end, fde_count, ip;
-  Elf_W(Addr) load_base, segbase = 0, max_load_addr = 0;
+  Elf_W(Addr) load_base, max_load_addr = 0;
   int ret, need_unwind_info = cb_data->need_unwind_info;
   unw_proc_info_t *pi = cb_data->pi;
   struct dwarf_eh_frame_hdr *hdr;
   unw_accessors_t *a;
   long n;
   int found = 0;
+#ifdef CONFIG_DEBUG_FRAME
+  unw_word_t start, end;
+#endif /* CONFIG_DEBUG_FRAME*/
 
   ip = cb_data->ip;
 
@@ -614,27 +586,6 @@ dwarf_callback (struct dl_phdr_info *info, size_t size, void *ptr)
 
   if (p_eh_hdr)
     {
-      if (likely (p_eh_hdr->p_vaddr >= p_text->p_vaddr
-		  && p_eh_hdr->p_vaddr < p_text->p_vaddr + p_text->p_memsz))
-	/* normal case: eh-hdr is inside text segment */
-	segbase = p_text->p_vaddr + load_base;
-      else
-	{
-	  /* Special case: eh-hdr is in some other segment; this may
-	     happen, e.g., for the Linux kernel's gate DSO, for
-	     example.  */
-	  phdr = info->dlpi_phdr;
-	  for (n = info->dlpi_phnum; --n >= 0; phdr++)
-	    {
-	      if (phdr->p_type == PT_LOAD && p_eh_hdr->p_vaddr >= phdr->p_vaddr
-		  && p_eh_hdr->p_vaddr < phdr->p_vaddr + phdr->p_memsz)
-		{
-		  segbase = phdr->p_vaddr + load_base;
-		  break;
-		}
-	    }
-	}
-
       if (p_dynamic)
 	{
 	  /* For dynamicly linked executables and shared libraries,
@@ -730,9 +681,29 @@ dwarf_callback (struct dl_phdr_info *info, size_t size, void *ptr)
     }
 
 #ifdef CONFIG_DEBUG_FRAME
-  {
-      found = dwarf_find_debug_frame (found, &cb_data->di_debug, info, ip);
-  }
+  /* Find the start/end of the described region by parsing the phdr_info
+     structure.  */
+  start = (unw_word_t) -1;
+  end = 0;
+
+  for (n = 0; n < info->dlpi_phnum; n++)
+    {
+      if (info->dlpi_phdr[n].p_type == PT_LOAD)
+        {
+	  unw_word_t seg_start = info->dlpi_addr + info->dlpi_phdr[n].p_vaddr;
+          unw_word_t seg_end = seg_start + info->dlpi_phdr[n].p_memsz;
+
+	  if (seg_start < start)
+	    start = seg_start;
+
+	  if (seg_end > end)
+	    end = seg_end;
+	}
+    }
+
+  found = dwarf_find_debug_frame (found, &cb_data->di_debug, ip,
+				  info->dlpi_addr, info->dlpi_name, start,
+				  end);
 #endif  /* CONFIG_DEBUG_FRAME */
 
   return found;
@@ -786,7 +757,7 @@ static inline const struct table_entry *
 lookup (const struct table_entry *table, size_t table_size, int32_t rel_ip)
 {
   unsigned long table_len = table_size / sizeof (struct table_entry);
-  const struct table_entry *e = 0;
+  const struct table_entry *e = NULL;
   unsigned long lo, hi, mid;
 
   /* do a binary search for right entry: */
@@ -794,7 +765,7 @@ lookup (const struct table_entry *table, size_t table_size, int32_t rel_ip)
     {
       mid = (lo + hi) / 2;
       e = table + mid;
-      Debug (1, "e->start_ip_offset = %lx\n", (long) e->start_ip_offset);
+      Debug (15, "e->start_ip_offset = %lx\n", (long) e->start_ip_offset);
       if (rel_ip < e->start_ip_offset)
 	hi = mid;
       else
