@@ -23,11 +23,6 @@ LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.  */
 
-#include <assert.h>
-#include <stdio.h>
-#include <stdlib.h>
-#include <string.h>
-
 #include "unwind_i.h"
 
 /* forward declaration: */
@@ -114,7 +109,7 @@ free_state_stack (struct ia64_reg_state *rs)
 
 /* Unwind decoder routines */
 
-static enum ia64_pregnum __attribute__ ((const))
+static enum ia64_pregnum CONST_ATTR
 decode_abreg (unsigned char abreg, int memory)
 {
   switch (abreg)
@@ -152,7 +147,7 @@ decode_abreg (unsigned char abreg, int memory)
     default:
       break;
     }
-  dprintf ("libunwind: bad abreg=0x%x\n", abreg);
+  Dprintf ("libunwind: bad abreg=0x%x\n", abreg);
   return IA64_REG_LC;
 }
 
@@ -198,7 +193,7 @@ spill_next_when (struct ia64_reg_info **regp, struct ia64_reg_info *lim,
 	  return;
 	}
     }
-  dprintf ("libunwind: excess spill!\n");
+  Dprintf ("libunwind: excess spill!\n");
 }
 
 static inline void
@@ -210,7 +205,7 @@ finish_prologue (struct ia64_state_record *sr)
 
   /* First, resolve implicit register save locations (see Section
      "11.4.2.3 Rules for Using Unwind Descriptors", rule 3). */
-  for (i = 0; i < (int) NELEMS (unw.save_order); ++i)
+  for (i = 0; i < (int) ARRAY_SIZE (unw.save_order); ++i)
     {
       reg = sr->curr.reg + unw.save_order[i];
       if (reg->where == IA64_WHERE_GR_SAVE)
@@ -694,6 +689,8 @@ desc_spill_sprel_p (unsigned char qp, unw_word t, unsigned char abreg,
 
 #include "unwind_decoder.h"
 
+#ifdef _U_dyn_op
+
 /* parse dynamic unwind info */
 
 static struct ia64_reg_info *
@@ -738,7 +735,7 @@ lookup_preg (int regnum, int memory, struct ia64_state_record *sr)
       break;
 
     default:
-      dprintf ("%s: invalid register number %d\n", __FUNCTION__, regnum);
+      Dprintf ("%s: invalid register number %d\n", __FUNCTION__, regnum);
       return NULL;
     }
   return sr->curr.reg + preg;
@@ -845,7 +842,7 @@ parse_dynamic (struct cursor *c, struct ia64_state_record *sr)
 		where = IA64_WHERE_BR;
 	      else
 		{
-		  dprintf ("%s: can't save to register number %d\n",
+		  Dprintf ("%s: can't save to register number %d\n",
 			   __FUNCTION__, (int) op->reg);
 		  return -UNW_EBADREG;
 		}
@@ -875,7 +872,7 @@ parse_dynamic (struct cursor *c, struct ia64_state_record *sr)
 		{
 		  if (val & 0xf)
 		    {
-		      dprintf ("%s: frame-size %ld not an integer "
+		      Dprintf ("%s: frame-size %ld not an integer "
 			       "multiple of 16\n",
 			       __FUNCTION__, (long) op->val);
 		      return -UNW_EINVAL;
@@ -884,7 +881,7 @@ parse_dynamic (struct cursor *c, struct ia64_state_record *sr)
 		}
 	      else
 		{
-		  dprintf ("%s: can only ADD to stack-pointer\n",
+		  Dprintf ("%s: can only ADD to stack-pointer\n",
 			   __FUNCTION__);
 		  return -UNW_EBADREG;
 		}
@@ -916,6 +913,9 @@ parse_dynamic (struct cursor *c, struct ia64_state_record *sr)
     }
   return 0;
 }
+#else
+# define parse_dynamic(c,sr)	(-UNW_EINVAL)
+#endif /* _U_dyn_op */
 
 
 HIDDEN int
@@ -1047,40 +1047,40 @@ create_state_record_for (struct cursor *c, struct ia64_state_record *sr,
 #if UNW_DEBUG
   if (unwi_debug_level > 2)
     {
-      dprintf ("%s: state record for func 0x%lx, t=%u (flags=0x%lx):\n",
+      Dprintf ("%s: state record for func 0x%lx, t=%u (flags=0x%lx):\n",
 	       __FUNCTION__,
 	       (long) c->pi.start_ip, sr->when_target, (long) c->pi.flags);
       for (r = sr->curr.reg; r < sr->curr.reg + IA64_NUM_PREGS; ++r)
 	{
 	  if (r->where != IA64_WHERE_NONE || r->when != IA64_WHEN_NEVER)
 	    {
-	      dprintf ("  %s <- ", unw.preg_name[r - sr->curr.reg]);
+	      Dprintf ("  %s <- ", unw.preg_name[r - sr->curr.reg]);
 	      switch (r->where)
 		{
 		case IA64_WHERE_GR:
-		  dprintf ("r%lu", (long) r->val);
+		  Dprintf ("r%lu", (long) r->val);
 		  break;
 		case IA64_WHERE_FR:
-		  dprintf ("f%lu", (long) r->val);
+		  Dprintf ("f%lu", (long) r->val);
 		  break;
 		case IA64_WHERE_BR:
-		  dprintf ("b%lu", (long) r->val);
+		  Dprintf ("b%lu", (long) r->val);
 		  break;
 		case IA64_WHERE_SPREL:
-		  dprintf ("[sp+0x%lx]", (long) r->val);
+		  Dprintf ("[sp+0x%lx]", (long) r->val);
 		  break;
 		case IA64_WHERE_PSPREL:
-		  dprintf ("[psp+0x%lx]", (long) r->val);
+		  Dprintf ("[psp+0x%lx]", (long) r->val);
 		  break;
 		case IA64_WHERE_NONE:
-		  dprintf ("%s+0x%lx",
+		  Dprintf ("%s+0x%lx",
 			   unw.preg_name[r - sr->curr.reg], (long) r->val);
 		  break;
 		default:
-		  dprintf ("BADWHERE(%d)", r->where);
+		  Dprintf ("BADWHERE(%d)", r->where);
 		  break;
 		}
-	      dprintf ("\t\t%d\n", r->when);
+	      Dprintf ("\t\t%d\n", r->when);
 	    }
 	}
     }
@@ -1117,9 +1117,15 @@ ia64_free_state_record (struct ia64_state_record *sr)
 HIDDEN int
 ia64_make_proc_info (struct cursor *c)
 {
-  if (c->as->caching_policy == UNW_CACHE_NONE
-      || ia64_get_cached_proc_info (c) < 0)
-    /* Lookup it up the slow way... */
-    return ia64_fetch_proc_info (c, c->ip, 0);
+  int ret, caching = c->as->caching_policy != UNW_CACHE_NONE;
+
+  if (!caching || ia64_get_cached_proc_info (c) < 0)
+    {
+      /* Lookup it up the slow way... */
+      if ((ret = ia64_fetch_proc_info (c, c->ip, 0)) < 0)
+	return ret;
+      if (caching)
+	ia64_cache_proc_info (c);
+    }
   return 0;
 }
