@@ -56,7 +56,7 @@ static __thread struct ia64_script_cache ia64_per_thread_cache =
   };
 #endif
 
-static inline unw_hash_index_t
+static inline unw_hash_index_t CONST_ATTR
 hash (unw_word_t ip)
 {
   /* based on (sqrt(5)/2-1)*2^64 */
@@ -113,11 +113,10 @@ get_script_cache (unw_addr_space_t as, intrmask_t *saved_maskp)
   if (AO_test_and_set (&cache->busy) == AO_TS_SET)
     return NULL;
 # else
-  sigprocmask (SIG_SETMASK, &unwi_full_mask, saved_maskp);
   if (likely (caching == UNW_CACHE_GLOBAL))
     {
-      Debug (16, "%s: acquiring lock\n", __FUNCTION__);
-      mutex_lock (&cache->lock);
+      Debug (16, "acquiring lock\n");
+      lock_acquire (&cache->lock, *saved_maskp);
     }
 # endif
 #endif
@@ -144,8 +143,7 @@ put_script_cache (unw_addr_space_t as, struct ia64_script_cache *cache,
   AO_CLEAR (&cache->busy);
 # else
   if (likely (as->caching_policy == UNW_CACHE_GLOBAL))
-    mutex_unlock (&cache->lock);
-  sigprocmask (SIG_SETMASK, saved_maskp, NULL);
+    lock_release (&cache->lock, *saved_maskp);
 # endif
 #endif
 }
