@@ -28,7 +28,7 @@ WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.  */
 PROTECTED _Unwind_Reason_Code
 _Unwind_RaiseException (struct _Unwind_Exception *exception_object)
 {
-  unsigned long exception_class = exception_object->exception_class;
+  uint64_t exception_class = exception_object->exception_class;
   _Unwind_Personality_Fn personality;
   struct _Unwind_Context context;
   _Unwind_Reason_Code reason;
@@ -60,7 +60,7 @@ _Unwind_RaiseException (struct _Unwind_Exception *exception_object)
       if (unw_get_proc_info (&context.cursor, &pi) < 0)
 	return _URC_FATAL_PHASE1_ERROR;
 
-      personality = (_Unwind_Personality_Fn) pi.handler;
+      personality = (_Unwind_Personality_Fn) (uintptr_t) pi.handler;
       if (personality)
 	{
 	  reason = (*personality) (_U_VERSION, _UA_SEARCH_PHASE,
