@@ -26,6 +26,10 @@ LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.  */
 
+#ifdef HAVE_CONFIG_H
+#include <config.h>
+#endif
+
 #include <stdlib.h>
 #include <string.h>
 #include <sys/mman.h>
@@ -43,47 +47,6 @@ static struct unw_addr_space local_addr_space;
 
 PROTECTED unw_addr_space_t unw_local_addr_space = &local_addr_space;
 
-static inline void *
-uc_addr (ucontext_t *uc, int reg)
-{
-  void *addr;
-
-  switch (reg)
-    {
-    case UNW_X86_64_R8: addr = &uc->uc_mcontext.gregs[REG_R8]; break;
-    case UNW_X86_64_R9: addr = &uc->uc_mcontext.gregs[REG_R9]; break;
-    case UNW_X86_64_R10: addr = &uc->uc_mcontext.gregs[REG_R10]; break;
-    case UNW_X86_64_R11: addr = &uc->uc_mcontext.gregs[REG_R11]; break;
-    case UNW_X86_64_R12: addr = &uc->uc_mcontext.gregs[REG_R12]; break;
-    case UNW_X86_64_R13: addr = &uc->uc_mcontext.gregs[REG_R13]; break;
-    case UNW_X86_64_R14: addr = &uc->uc_mcontext.gregs[REG_R14]; break;
-    case UNW_X86_64_R15: addr = &uc->uc_mcontext.gregs[REG_R15]; break;
-    case UNW_X86_64_RDI: addr = &uc->uc_mcontext.gregs[REG_RDI]; break;
-    case UNW_X86_64_RSI: addr = &uc->uc_mcontext.gregs[REG_RSI]; break;
-    case UNW_X86_64_RBP: addr = &uc->uc_mcontext.gregs[REG_RBP]; break;
-    case UNW_X86_64_RBX: addr = &uc->uc_mcontext.gregs[REG_RBX]; break;
-    case UNW_X86_64_RDX: addr = &uc->uc_mcontext.gregs[REG_RDX]; break;
-    case UNW_X86_64_RAX: addr = &uc->uc_mcontext.gregs[REG_RAX]; break;
-    case UNW_X86_64_RCX: addr = &uc->uc_mcontext.gregs[REG_RCX]; break;
-    case UNW_X86_64_RSP: addr = &uc->uc_mcontext.gregs[REG_RSP]; break;
-    case UNW_X86_64_RIP: addr = &uc->uc_mcontext.gregs[REG_RIP]; break;
-
-    default:
-      addr = NULL;
-    }
-  return addr;
-}
-
-# ifdef UNW_LOCAL_ONLY
-
-HIDDEN void *
-tdep_uc_addr (ucontext_t *uc, int reg)
-{
-  return uc_addr (uc, reg);
-}
-
-# endif /* UNW_LOCAL_ONLY */
-
 HIDDEN unw_dyn_info_list_t _U_dyn_info_list;
 
 /* XXX fix me: there is currently no way to locate the dyn-info list
@@ -108,6 +71,42 @@ get_dyn_info_list_addr (unw_addr_space_t as, unw_word_t *dyn_info_list_addr,
 #define PAGE_SIZE 4096
 #define PAGE_START(a)	((a) & ~(PAGE_SIZE-1))
 
+static int (*mem_validate_func) (void *addr, size_t len);
+static int msync_validate (void *addr, size_t len)
+{
+  return msync (addr, len, MS_ASYNC);
+}
+
+#ifdef HAVE_MINCORE
+static int mincore_validate (void *addr, size_t len)
+{
+  unsigned char mvec[2]; /* Unaligned access may cross page boundary */
+  return mincore (addr, len, mvec);
+}
+#endif
+
+/* Initialise memory validation method. On linux kernels <2.6.21,
+   mincore() returns incorrect value for MAP_PRIVATE mappings,
+   such as stacks. If mincore() was available at compile time,
+   check if we can actually use it. If not, use msync() instead. */
+HIDDEN void
+tdep_init_mem_validate (void)
+{
+#ifdef HAVE_MINCORE
+  unsigned char present = 1;
+  if (mincore (&present, 1, &present) == 0)
+    {
+      Debug(1, "using mincore to validate memory\n");
+      mem_validate_func = mincore_validate;
+    }
+  else
+#endif
+    {
+      Debug(1, "using msync to validate memory\n");
+      mem_validate_func = msync_validate;
+    }
+}
+
 /* Cache of already validated addresses */
 #define NLGA 4
 static unw_word_t last_good_addr[NLGA];
@@ -117,16 +116,25 @@ static int
 validate_mem (unw_word_t addr)
 {
   int i, victim;
+  size_t len;
+
+  if (PAGE_START(addr + sizeof (unw_word_t) - 1) == PAGE_START(addr))
+    len = PAGE_SIZE;
+  else
+    len = PAGE_SIZE * 2;
 
   addr = PAGE_START(addr);
 
+  if (addr == 0)
+    return -1;
+
   for (i = 0; i < NLGA; i++)
     {
       if (last_good_addr[i] && (addr == last_good_addr[i]))
 	return 0;
     }
 
-  if (msync ((void *) addr, 1, MS_SYNC) == -1)
+  if (mem_validate_func ((void *) addr, len) == -1)
     return -1;
 
   victim = lga_victim;
@@ -150,7 +158,7 @@ static int
 access_mem (unw_addr_space_t as, unw_word_t addr, unw_word_t *val, int write,
 	    void *arg)
 {
-  if (write)
+  if (unlikely (write))
     {
       Debug (16, "mem[%016lx] <- %lx\n", addr, *val);
       *(unw_word_t *) addr = *val;
@@ -159,7 +167,8 @@ access_mem (unw_addr_space_t as, unw_word_t addr, unw_word_t *val, int write,
     {
       /* validate address */
       const struct cursor *c = (const struct cursor *)arg;
-      if (c && c->validate && validate_mem(addr))
+      if (likely (c != NULL) && unlikely (c->validate)
+          && unlikely (validate_mem (addr)))
         return -1;
       *val = *(unw_word_t *) addr;
       Debug (16, "mem[%016lx] -> %lx\n", addr, *val);
@@ -177,7 +186,7 @@ access_reg (unw_addr_space_t as, unw_regnum_t reg, unw_word_t *val, int write,
   if (unw_is_fpreg (reg))
     goto badreg;
 
-  if (!(addr = uc_addr (uc, reg)))
+  if (!(addr = x86_64_r_uc_addr (uc, reg)))
     goto badreg;
 
   if (write)
@@ -207,7 +216,7 @@ access_fpreg (unw_addr_space_t as, unw_regnum_t reg, unw_fpreg_t *val,
   if (!unw_is_fpreg (reg))
     goto badreg;
 
-  if (!(addr = uc_addr (uc, reg)))
+  if (!(addr = x86_64_r_uc_addr (uc, reg)))
     goto badreg;
 
   if (write)
