@@ -43,7 +43,7 @@ unw_init_local (unw_cursor_t *cursor, ucontext_t *uc)
 {
   struct cursor *c = (struct cursor *) cursor;
 
-  if (tdep_needs_initialization)
+  if (unlikely (!tdep_init_done))
     tdep_init ();
 
   Debug (1, "(cursor=%p)\n", c);
@@ -52,7 +52,7 @@ unw_init_local (unw_cursor_t *cursor, ucontext_t *uc)
   c->dwarf.as_arg = c;
   c->uc = uc;
   c->validate = 0;
-  return common_init (c);
+  return common_init (c, 1);
 }
 
 #endif /* !UNW_REMOTE_ONLY */
