@@ -36,7 +36,7 @@ unw_init_remote (unw_cursor_t *cursor, unw_addr_space_t as, void *as_arg)
 #else /* !UNW_LOCAL_ONLY */
   struct cursor *c = (struct cursor *) cursor;
 
-  if (tdep_needs_initialization)
+  if (!tdep_init_done)
     tdep_init ();
 
   Debug (1, "(cursor=%p)\n", c);
@@ -50,8 +50,8 @@ unw_init_remote (unw_cursor_t *cursor, unw_addr_space_t as, void *as_arg)
   else
     {
       c->dwarf.as_arg = as_arg;
-      c->uc = 0;
+      c->uc = NULL;
     }
-  return common_init (c);
+  return common_init (c, 0);
 #endif /* !UNW_LOCAL_ONLY */
 }
