@@ -1,5 +1,5 @@
 /* libunwind - a platform-independent unwind library
-   Copyright (C) 2003 Hewlett-Packard Co
+   Copyright (C) 2003-2004 Hewlett-Packard Co
 	Contributed by David Mosberger-Tang <davidm@hpl.hp.com>
 
 Permission is hereby granted, free of charge, to any person obtaining
@@ -27,11 +27,15 @@ WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.  */
 # include "config.h"
 #endif
 
+#include "compiler.h"
+
 #include <libunwind.h>
 #include <signal.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <string.h>
 #include <unistd.h>
+#include <errno.h>
 
 #ifdef HAVE_IA64INTRIN_H
 # include <ia64intrin.h>
@@ -45,23 +49,40 @@ int nerrors;
 int got_usr1, got_usr2;
 char *sigusr1_sp;
 
+uintptr_t
+get_bsp (void)
+{
+#if UNW_TARGET_IA64
+# ifdef __INTEL_COMPILER
+  return __getReg (_IA64_REG_AR_BSP);
+# else
+  return (uintptr_t) __builtin_ia64_bsp ();
+# endif
+#else
+  return 0;
+#endif
+}
+
+#ifdef TEST_WITH_SIGINFO
+void
+handler (int sig,
+         siginfo_t *si UNUSED,
+         void *ucontext UNUSED)
+#else
 void
 handler (int sig)
+#endif
 {
   unw_word_t ip;
-  sigset_t mask;
+  sigset_t mask, oldmask;
   unw_context_t uc;
   unw_cursor_t c;
   char foo;
+  int ret;
 
 #if UNW_TARGET_IA64
-# ifdef __ECC
-  void *bsp = (void *) __getReg(_IA64_REG_AR_BSP);
-# else
-  void *bsp = __builtin_ia64_bsp ();
-#endif
   if (verbose)
-    printf ("bsp = %p\n", bsp);
+    printf ("bsp = %llx\n", (unsigned long long) get_bsp ());
 #endif
 
   if (verbose)
@@ -74,30 +95,39 @@ handler (int sig)
 
       sigemptyset (&mask);
       sigaddset (&mask, SIGUSR2);
-      sigprocmask (SIG_BLOCK, &mask, NULL);
+      sigprocmask (SIG_BLOCK, &mask, &oldmask);
       kill (getpid (), SIGUSR2);	/* pend SIGUSR2 */
 
       signal (SIGUSR1, SIG_IGN);
-      signal (SIGUSR2, handler);
 
-      unw_getcontext(&uc);
-      unw_init_local(&c, &uc);
-      unw_step(&c);		/* step to signal trampoline */
-      unw_step(&c);		/* step to signaller frame (main ()) */
-      unw_get_reg(&c, UNW_REG_IP, &ip);
+      if ((ret = unw_getcontext (&uc)) < 0)
+	panic ("unw_getcontext() failed: ret=%d\n", ret);
+#if UNW_TARGET_X86_64
+      /* unw_getcontext() doesn't save signal mask to avoid a syscall */
+      uc.uc_sigmask = oldmask; 
+#endif
+      if ((ret = unw_init_local (&c, &uc)) < 0)
+	panic ("unw_init_local() failed: ret=%d\n", ret);
+
+      if ((ret = unw_step (&c)) < 0)		/* step to signal trampoline */
+	panic ("unw_step(1) failed: ret=%d\n", ret);
+
+      if ((ret = unw_step (&c)) < 0)		/* step to kill() */
+	panic ("unw_step(2) failed: ret=%d\n", ret);
+
+      if ((ret = unw_get_reg (&c, UNW_REG_IP, &ip)) < 0)
+	panic ("unw_get_reg(IP) failed: ret=%d\n", ret);
       if (verbose)
-	printf ("resuming at 0x%lx, with SIGUSR2 pending\n", ip);
-      unw_resume(&c);
+	printf ("resuming at 0x%lx, with SIGUSR2 pending\n",
+		(unsigned long) ip);
+      unw_resume (&c);
     }
   else if (sig == SIGUSR2)
     {
       ++got_usr2;
       if (got_usr1)
 	{
-	  if (sigusr1_sp != &foo)
-	    panic ("Stack pointer changed from %p to %p between signals\n",
-		   sigusr1_sp, &foo);
-	  else if (verbose)
+	  if (verbose)
 	    printf ("OK: stack still at %p\n", &foo);
 	}
       signal (SIGUSR2, SIG_IGN);
@@ -107,19 +137,53 @@ handler (int sig)
 }
 
 int
-main (int argc, char **argv)
+main (int argc, char **argv UNUSED)
 {
+  struct sigaction sa;
+  float d = 1.0;
+  int n = 0;
+
   if (argc > 1)
     verbose = 1;
 
-  signal (SIGUSR1, handler);
+  memset (&sa, 0, sizeof(sa));
+#ifdef TEST_WITH_SIGINFO
+  sa.sa_sigaction = handler;
+  sa.sa_flags = SA_SIGINFO;
+#else
+  sa.sa_handler = handler;
+#endif
+
+  if (sigaction (SIGUSR1, &sa, NULL) != 0 ||
+      sigaction (SIGUSR2, &sa, NULL) != 0)
+    {
+      fprintf (stderr, "sigaction() failed: %s\n", strerror (errno));
+      return -1;
+    }
+
+  /* Use the FPU a bit; otherwise we get spurious errors should the
+     signal handler need to use the FPU for any reason.  This seems to
+     happen on x86-64.  */
+  while (d > 0.0)
+    {
+      d /= 2.0;
+      ++n;
+    }
+  if (n > 9999)
+    return -1;	/* can't happen, but don't tell the compiler... */
 
   if (verbose)
     printf ("sending SIGUSR1\n");
   kill (getpid (), SIGUSR1);
 
+  if (!got_usr2)
+    panic ("failed to get SIGUSR2\n");
+
   if (nerrors)
-    fprintf (stderr, "FAILURE: detected %d errors\n", nerrors);
+    {
+      fprintf (stderr, "FAILURE: detected %d errors\n", nerrors);
+      exit (-1);
+    }
 
   if (verbose)
     printf ("SUCCESS\n");
